#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_measure_particle_number_on_one_mode():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 2)) * 3j
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(1, 0, 1)) * (-3j)

        pq.Q() | pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 1)) * 1j
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 1)) * (-1j)

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 1, 1)) * 2j
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 1)) * (-2j)

        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) * 2 / 6
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) * 1 / 6
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) * 3 / 6

        pq.Q(2) | pq.ParticleNumberMeasurement()

    config = pq.Config(cutoff=3)

    simulator = pq.FockSimulator(d=3, config=config)

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (1,) or sample == (2,)

    if sample == (1,):
        expected_simulator = pq.FockSimulator(
            d=3,
            config=config,
        )
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                1 / 3 * pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)),
                4j * pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 1, 1)),
                -2j * pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 1)),
                -4j * pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 1)),
                2 / 3 * pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)),
                2j * pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 1)),
            ]
        ).state

    elif sample == (2,):
        expected_simulator = pq.FockSimulator(
            d=3,
            config=config,
        )
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2))]
        ).state

    assert result.state == expected_state


def test_measure_particle_number_on_two_modes():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 2)) * 3j
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(1, 0, 1)) * (-3j)

        pq.Q() | pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 1)) * 1j
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 1)) * (-1j)

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 1, 1)) * 2j
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 1)) * (-2j)

        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) * 2 / 6
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) * 1 / 6
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) * 3 / 6

        pq.Q(1, 2) | pq.ParticleNumberMeasurement()

    config = pq.Config(cutoff=3)

    simulator = pq.FockSimulator(d=3, config=config)

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (0, 1) or sample == (1, 1) or sample == (0, 2)

    if sample == (0, 1):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)),
                pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 1)) * (-6j),
                pq.DensityMatrix(ket=(1, 0, 1), bra=(0, 0, 1)) * 6j,
            ]
        ).state

    elif sample == (1, 1):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)),
            ]
        ).state

    elif sample == (0, 2):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2))]
        ).state

    assert result.state == expected_state


def test_measure_particle_number_on_all_modes():
    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 0), bra=(0, 0, 0)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 0), bra=(1, 0, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 0), bra=(0, 0, 1)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q() | pq.ParticleNumberMeasurement()

    config = pq.Config(cutoff=3)

    simulator = pq.FockSimulator(d=3, config=config)

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (0, 0, 0) or sample == (0, 0, 1) or sample == (1, 0, 0)

    if sample == (0, 0, 0):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.DensityMatrix(ket=(0, 0, 0), bra=(0, 0, 0)),
            ]
        ).state

    elif sample == (0, 0, 1):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)),
            ]
        ).state

    elif sample == (1, 0, 0):
        expected_simulator = pq.FockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.DensityMatrix(ket=(1, 0, 0), bra=(1, 0, 0)),
            ]
        ).state

    assert result.state == expected_state


def test_measure_particle_number_with_multiple_shots():
    shots = 4

    with pq.Program() as preparation:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 0), bra=(0, 0, 0)) / 4

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 0), bra=(1, 0, 0)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(1, 0, 0)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(1, 0, 0), bra=(0, 0, 1)) * np.sqrt(1 / 8)

    with pq.Program() as program:
        pq.Q() | preparation

        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    result = simulator.execute(program, shots)

    assert np.isclose(sum(result.state.fock_probabilities), 1)
    assert len(result.samples) == shots



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from scipy.stats import unitary_group
from scipy.linalg import block_diag


def test_Displacement_mean_photon_number_gradient_1_mode(tf):
    r = tf.Variable(0.43)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=6), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r)

        state = simulator.execute(program).state

        mean = state.mean_photon_number()

    gradient = tape.gradient(mean, [r])

    assert np.isclose(mean, r**2)
    assert np.isclose(gradient, 2 * r)


def test_Displacement_mean_photon_number_gradient_1_mode_with_phaseshift(tf):
    r = 0.1
    phi = tf.Variable(np.pi / 5)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=8), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r, phi=phi)

        state = simulator.execute(program).state

        mean, _ = state.quadratures_mean_variance(modes=(0,))

    gradient = tape.gradient(mean, [phi])

    assert np.isclose(mean, np.sqrt(2 * simulator.config.hbar) * r * np.cos(phi))
    assert np.isclose(gradient, -np.sqrt(2 * simulator.config.hbar) * r * np.sin(phi))


def test_Displacement_mean_photon_number_gradient_2_modes(tf):
    r = tf.Variable([0.15, 0.2])

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=6),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r[0])
            pq.Q(1) | pq.Displacement(r=r[1])

        result = simulator.execute(program)

        state = result.state

        mean = state.mean_photon_number()

    gradient = tape.gradient(mean, [r])

    assert np.isclose(mean, sum(r**2))
    assert np.allclose(gradient, 2 * r)


def test_Displacement_fock_probabilities_jacobian_2_modes(tf):
    r = tf.Variable([0.15, 0.2])

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r[0])
            pq.Q(1) | pq.Displacement(r=r[1])

        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [r])

    expected_fock_probabilities = np.exp(-sum(r**2)) * np.array(
        [
            1.0,
            r[0] ** 2,
            r[1] ** 2,
            r[0] ** 4 / 2,
            r[0] ** 2 * r[1] ** 2,
            r[1] ** 4 / 2,
        ]
    )

    expected_jacobian = np.exp(-sum(r**2)) * np.array(
        [
            -2 * r,
            [
                2 * r[0] * (1 - r[0] ** 2),
                -2 * r[1] * r[0] ** 2,
            ],
            [
                -2 * r[0] * r[1] ** 2,
                2 * r[1] * (1 - r[1] ** 2),
            ],
            [
                2 * r[0] ** 3 - r[0] ** 5,
                -2 * r[1] * r[0] ** 4 / 2,
            ],
            [
                r[1] ** 2 * 2 * r[0] * (1 - r[0] ** 2),
                r[0] ** 2 * 2 * r[1] * (1 - r[1] ** 2),
            ],
            [
                -2 * r[0] * r[1] ** 4 / 2,
                2 * r[1] ** 3 - r[1] ** 5,
            ],
        ]
    )

    assert np.allclose(
        fock_probabilities,
        expected_fock_probabilities,
        atol=1e-4,
    )

    assert np.allclose(
        jacobian,
        expected_jacobian,
        rtol=1e-2,
    )


def test_Squeezing_mean_photon_number_gradient_1_mode(tf):
    r = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=8),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r=r)

        state = simulator.execute(program).state

        mean = state.mean_photon_number()

    gradient = tape.gradient(mean, [r])

    assert np.isclose(mean, np.sinh(r) ** 2)
    assert np.isclose(gradient, 2 * np.sinh(r) * np.cosh(r))


def test_Squeezing_mean_photon_number_gradient_1_mode_with_phaseshift(tf):
    r = 0.1
    phi = tf.Variable(np.pi / 5)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=8),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r=r, phi=phi)

        state = simulator.execute(program).state

        _, variance = state.quadratures_mean_variance(modes=(0,))

    gradient = tape.jacobian(variance, [phi])

    assert np.isclose(variance, np.cosh(2 * r) - np.sinh(2 * r) * np.cos(phi))
    assert np.isclose(gradient, np.sinh(2 * r) * np.sin(phi))


def test_Beamsplitter_fock_probabilities_gradient_1_particle(tf):
    theta = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=2),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((1, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta)

        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [theta])

    assert np.allclose(
        fock_probabilities,
        [0, np.cos(theta) ** 2, np.sin(theta) ** 2],
    )

    assert np.allclose(
        jacobian,
        [0, -2 * np.cos(theta) * np.sin(theta), 2 * np.cos(theta) * np.sin(theta)],
    )


def test_Beamsplitter_fock_probabilities_gradient_1_particle_with_phaseshift(tf):
    theta = tf.Variable(np.pi / 3)
    phi = np.pi / 5

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=2),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((1, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta, phi=phi)

        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [theta])

    assert np.allclose(
        fock_probabilities,
        [0, np.cos(theta) ** 2, np.sin(theta) ** 2],
    )

    assert np.allclose(
        jacobian,
        [0, -2 * np.cos(theta) * np.sin(theta), 2 * np.cos(theta) * np.sin(theta)],
    )


def test_decomposed_Beamsplitter_state_vector_gradient_1_particle(tf):
    theta = tf.Variable(11 * np.pi / 17)

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=2),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((1, 0))

            pq.Q(1) | pq.Phaseshifter(-np.pi / 2)
            pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=0.0)
            pq.Q(0) | pq.Phaseshifter(theta)
            pq.Q(1) | pq.Phaseshifter(np.pi - theta)
            pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=0.0)
            pq.Q(1) | pq.Phaseshifter(-np.pi / 2)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [theta])

    assert np.allclose(
        state_vector,
        [0, np.cos(theta), np.sin(theta)],
    )

    assert np.allclose(
        jacobian,
        np.real([0, -np.sin(theta), np.cos(theta)]),
    )


def test_Beamsplitter_fock_probabilities_gradient_2_particles(tf):
    theta = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta)

        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [theta])

    assert np.allclose(
        fock_probabilities,
        [
            0,
            0,
            0,
            np.cos(theta) ** 4,
            2 * (np.cos(theta) * np.sin(theta)) ** 2,
            np.sin(theta) ** 4,
        ],
    )

    assert np.allclose(
        jacobian,
        [
            0,
            0,
            0,
            -4 * np.cos(theta) ** 3 * np.sin(theta),
            2 * (np.sin(2 * theta)) * np.cos(2 * theta),
            4 * np.sin(theta) ** 3 * np.cos(theta),
        ],
    )


def test_multiple_Beamsplitter_state_vector_gradient_2_particles(tf):
    theta_1 = np.pi / 7
    phi_1 = 5 * np.pi / 9
    theta = tf.Variable(np.pi / 5)

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta_1, phi=phi_1)

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta, phi=0.0)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [theta])

    u00 = np.cos(theta_1) * np.cos(theta) - np.exp(1j * phi_1) * np.sin(
        theta_1
    ) * np.sin(theta)

    u10 = np.cos(theta_1) * np.sin(theta) + np.exp(1j * phi_1) * np.cos(theta) * np.sin(
        theta_1
    )

    u00_diff = -u10
    u10_diff = u00

    assert np.allclose(
        state_vector,
        [
            0,
            0,
            0,
            u00**2,
            np.sqrt(2) * u00 * u10,
            u10**2,
        ],
    )

    assert np.allclose(
        jacobian,
        np.real(
            [
                0,
                0,
                0,
                2 * u00 * u00_diff,
                np.sqrt(2) * (u00_diff * u10 + u00 * u10_diff),
                2 * u10 * u10_diff,
            ]
        ),
    )


def test_multiple_Beamsplitter_state_vector_gradient_2_particles_reversed(tf):
    theta_1 = np.pi / 7
    phi_1 = 5 * np.pi / 9
    theta = tf.Variable(np.pi / 5)

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta, phi=0.0)

            pq.Q(0, 1) | pq.Beamsplitter(theta=theta_1, phi=phi_1)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [theta])

    u00 = np.cos(theta_1) * np.cos(theta) - np.exp(-1j * phi_1) * np.sin(
        theta_1
    ) * np.sin(theta)

    u10 = np.cos(theta_1) * np.sin(theta) + np.exp(1j * phi_1) * np.cos(theta) * np.sin(
        theta_1
    )

    u00_diff = -u10.conj()
    u10_diff = u00.conj()

    assert np.allclose(
        state_vector,
        [
            0,
            0,
            0,
            u00**2,
            np.sqrt(2) * u00 * u10,
            u10**2,
        ],
    )

    assert np.allclose(
        jacobian,
        np.real(
            [
                0,
                0,
                0,
                2 * u00 * u00_diff,
                np.sqrt(2) * (u00_diff * u10 + u00 * u10_diff),
                2 * u10 * u10_diff,
            ]
        ),
    )


@pytest.mark.monkey
def test_jacobian_of_state_after_mixing_with_fix_Interferometer(tf):
    c1 = tf.Variable(np.sqrt(0.2))
    c2 = tf.Variable(np.sqrt(0.3))
    c3 = tf.Variable(np.sqrt(0.5))

    d = 3

    simulator = pq.PureFockSimulator(
        d=d,
        config=pq.Config(cutoff=2),
        connector=pq.TensorflowConnector(),
    )

    U = unitary_group.rvs(3)

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((1, 0, 0)) * c1
            pq.Q(all) | pq.StateVector((0, 1, 0)) * c2
            pq.Q(all) | pq.StateVector((0, 0, 1)) * c3

            pq.Q(all) | pq.Interferometer(U)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [c1, c2, c3])

    initial_state_vector = np.array([0.0, c1, c2, c3])

    fock_space_unitary = block_diag([1.0], U)

    assert np.allclose(state_vector, fock_space_unitary @ initial_state_vector)

    assert np.allclose(
        jacobian,
        np.real(
            [
                fock_space_unitary @ np.array([0, 1, 0, 0]),
                fock_space_unitary @ np.array([0, 0, 1, 0]),
                fock_space_unitary @ np.array([0, 0, 0, 1]),
            ]
        ),
    )


def test_Phaseshifter_density_matrix_gradient(tf):
    phi = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=2), connector=pq.TensorflowConnector()
    )

    coefficients = [np.sqrt(0.6), np.sqrt(0.4)]

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(0) | pq.StateVector([0]) * coefficients[0]
            pq.Q(0) | pq.StateVector([1]) * coefficients[1]

            pq.Q(0) | pq.Phaseshifter(phi)

        state = simulator.execute(program).state

        density_matrix = state.density_matrix

    jacobian = tape.jacobian(density_matrix, [phi])

    assert np.allclose(
        density_matrix,
        [
            [
                coefficients[0] ** 2,
                coefficients[0] * coefficients[1] * np.exp(-1j * phi),
            ],
            [
                coefficients[0] * coefficients[1] * np.exp(1j * phi),
                coefficients[1] ** 2,
            ],
        ],
    )

    assert np.allclose(
        jacobian,
        [
            [0.0, -np.prod(coefficients) * np.sin(phi)],
            [-np.prod(coefficients) * np.sin(phi), 0.0],
        ],
    )


def test_Phaseshifter_density_matrix_gradient_is_zero_at_zero_phaseshift(tf):
    phi = tf.Variable(0.0)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(0) | pq.StateVector([0]) * np.sqrt(0.5)
            pq.Q(0) | pq.StateVector([1]) * np.sqrt(0.3)
            pq.Q(0) | pq.StateVector([2]) * np.sqrt(0.2)

            pq.Q(0) | pq.Phaseshifter(phi)

        state = simulator.execute(program).state

        density_matrix = state.density_matrix

    jacobian = tape.jacobian(density_matrix, [phi])

    assert np.allclose(jacobian, np.zeros_like(jacobian))


def test_Interferometer_fock_probabilities(tf):
    param = tf.Variable(0.0)

    config = pq.Config(cutoff=3)

    simulator = pq.PureFockSimulator(
        d=2, config=config, connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        hamiltonian = tf.stack(
            [
                tf.stack([0, tf.exp(param)]),
                tf.stack([tf.exp(param), 0]),
            ]
        )

        j = tf.complex(0.0, 1.0)
        interferometer = tf.linalg.expm(j * hamiltonian)
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((1, 0)) / np.sqrt(2)
            pq.Q(all) | pq.StateVector((1, 1)) / np.sqrt(2)

            pq.Q(all) | pq.Interferometer(interferometer)

        state = simulator.execute(program).state

        density_matrix = state.density_matrix

    jacobian = tape.jacobian(density_matrix, [param])

    expected_jacobian = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, -0.4546487, 0.0, 0.0, -0.3162077, 0.0],
            [0.0, 0.0, 0.4546488, -0.07391234, 0.0, -0.07391226],
            [0.0, 0.0, -0.07391234, -0.3784013, 0.0, -0.3784013],
            [0.0, -0.3162077, 0.0, 0.0, 0.75680256, 0.0],
            [0.0, 0.0, -0.07391226, -0.3784013, 0.0, -0.37840137],
        ],
    )

    assert np.allclose(jacobian, expected_jacobian)


def test_Squeezing2_mean_photon_number(tf):
    r = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(all) | pq.Vacuum()

            pq.Q(all) | pq.Squeezing2(r=r)

        state = simulator.execute(program).state

        state.normalize()

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [r])

    assert np.allclose(jacobian, [-0.19349256, 0.0, 0.0, 0.0, 0.19349256, 0.0])


def test_Kerr_fock_probabilities_on_1_mode(tf):
    xi = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=3), connector=pq.TensorflowConnector()
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0]) / np.sqrt(2)
        pq.Q(all) | pq.StateVector([2]) / np.sqrt(2)

        pq.Q(all) | pq.Kerr(xi=xi)

    with tf.GradientTape() as tape:
        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [xi])

    assert np.allclose(jacobian, [0.0, 0.0, 0.0], atol=1e-6)


def test_Kerr_density_matrix_on_1_mode(tf):
    xi = tf.Variable(np.pi / 5)

    n = 2

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=n + 1), connector=pq.TensorflowConnector()
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0]) / np.sqrt(2)
        pq.Q(all) | pq.StateVector([n]) / np.sqrt(2)

        pq.Q(all) | pq.Kerr(xi=xi)

    with tf.GradientTape() as tape:
        state = simulator.execute(program).state

        density_matrix = state.density_matrix

    jacobian = tape.jacobian(density_matrix, [xi])

    coefficient = n**2

    gradient_of_0n_component = 1j * coefficient * np.exp(1j * xi * coefficient) / 2

    excepted_jacobian = [
        [0.0, 0.0, np.conj(gradient_of_0n_component)],
        [0.0, 0.0, 0.0],
        [gradient_of_0n_component, 0.0, 0.0],
    ]

    assert np.allclose(jacobian, np.real(excepted_jacobian))


def test_CubicPhase_fock_probabilities_on_1_mode(tf):
    gamma = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=4), connector=pq.TensorflowConnector()
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([0]) / np.sqrt(2)
        pq.Q(all) | pq.StateVector([2]) / np.sqrt(2)

        pq.Q(all) | pq.CubicPhase(gamma=gamma)

    with tf.GradientTape() as tape:
        state = simulator.execute(program).state

        fock_probabilities = state.fock_probabilities

    jacobian = tape.jacobian(fock_probabilities, [gamma])

    assert np.allclose(jacobian, [-0.16857366, 0.35571513, -0.5196387, 0.33249724])


def test_CrossKerr_density_matrix(tf):
    xi = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=4), connector=pq.TensorflowConnector()
    )

    n = np.array(
        [
            [1, 1],
            [1, 2],
        ]
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q(0, 1) | pq.StateVector(n[0]) * np.sqrt(0.5)
            pq.Q(0, 1) | pq.StateVector(n[1]) * np.sqrt(0.5)

            pq.Q(all) | pq.CrossKerr(xi=xi)

        state = simulator.execute(program).state

        density_matrix = state.density_matrix

    jacobian = tape.jacobian(density_matrix, [xi])

    expected_jacobian = np.zeros_like(density_matrix)

    relative_phase_angle = n[0, 0] * n[0, 1] - n[1, 0] * n[1, 1]

    coefficient = 1j * relative_phase_angle * np.exp(1j * xi * relative_phase_angle)

    expected_jacobian[4, 8] = np.real(coefficient / 2)
    expected_jacobian[8, 4] = np.real(coefficient / 2)

    assert np.allclose(jacobian, expected_jacobian)


def test_mean_position_Displacement_gradient_on_1_mode(tf):
    d = 1
    cutoff = 7

    alpha_ = tf.Variable(0.02)

    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()

        pq.Q(all) | pq.Displacement(r=alpha_)

    config = pq.Config(cutoff=cutoff)

    simulator = pq.PureFockSimulator(
        d=d, config=config, connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        state = simulator.execute(program).state
        mean = state.mean_position(mode=0)

    grad = tape.gradient(mean, [alpha_])

    assert np.allclose(mean, np.sqrt(2 * config.hbar) * alpha_)
    assert np.allclose(grad, np.sqrt(2 * config.hbar))


def test_mean_position_Displacement_and_Squeezing_gradient_on_1_mode(tf):
    d = 1
    cutoff = 7

    alpha_ = tf.Variable(0.01)
    r_ = tf.Variable(0.05)

    with pq.Program() as program:
        pq.Q(all) | pq.Vacuum()

        pq.Q(all) | pq.Displacement(r=alpha_)
        pq.Q(all) | pq.Squeezing(r_)

    simulator = pq.PureFockSimulator(
        d=d, config=pq.Config(cutoff=cutoff), connector=pq.TensorflowConnector()
    )

    with tf.GradientTape() as tape:
        state = simulator.execute(program).state
        mean = state.mean_position(mode=0)

    grad = tape.gradient(mean, [r_, alpha_])

    assert np.allclose(mean, 0.019024584947048066)
    assert np.allclose(grad, [-0.019024614, 1.9024585382680494])


def test_Displacement_state_vector_gradient(tf):
    r = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [r])

    expected_state_vector = np.exp(-(r**2) / 2) * np.array([1, r, r**2 / np.sqrt(2)])
    expected_jacobian = np.array(
        [
            -r * np.exp(-(r**2) / 2),
            np.exp(-(r**2) / 2) * (1 - r**2),
            np.exp(-(r**2) / 2) * r * (2 - r**2) / np.sqrt(2),
        ]
    )
    assert np.allclose(state_vector, expected_state_vector)
    assert np.allclose(jacobian, expected_jacobian)


def test_complex_Displacement_state_vector_gradient(tf):
    r = tf.Variable(0.1)
    phi = tf.Variable(np.pi / 3)

    alpha = tf.cast(r, tf.complex64) * tf.exp(1j * tf.cast(phi, tf.complex64))

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=r, phi=phi)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [r, phi])

    expected_state_vector = np.exp(-(r**2) / 2) * np.array(
        [1, alpha, alpha**2 / np.sqrt(2)], dtype=complex
    )
    expected_jacobian = np.real(
        np.array(
            [
                [
                    -r * np.exp(-(r**2) / 2),
                    np.exp(-(r**2) / 2) * (1 - r**2) * np.exp(1j * phi),
                    np.exp(-(r**2) / 2)
                    * r
                    * (2 - r**2)
                    / np.sqrt(2)
                    * np.exp(1j * 2 * phi),
                ],
                [
                    0.0,
                    np.exp(-(r**2) / 2) * alpha * 1j,
                    np.exp(-(r**2) / 2) * alpha**2 / np.sqrt(2) * 2j,
                ],
            ]
        )
    )
    assert np.allclose(state_vector, expected_state_vector)
    assert np.allclose(jacobian, expected_jacobian)


def test_Squeezing_state_vector_gradient(tf):
    r = tf.Variable(0.1)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r=r)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [r])

    expected_state_vector = (
        1 / np.sqrt(np.cosh(r)) * np.array([1, 0, -np.tanh(r) * np.sqrt(2) / 2])
    )
    expected_jacobian = np.array(
        [
            -np.sinh(r) / (2 * np.cosh(r) ** (3 / 2)),
            0,
            (3 * np.sinh(r) ** 2 - 2 * np.cosh(r) ** 2)
            / (2 * np.sqrt(2) * np.cosh(r) ** (5 / 2)),
        ]
    )
    assert np.allclose(state_vector, expected_state_vector)
    assert np.allclose(jacobian, expected_jacobian)


def test_complex_Squeezing_state_vector_gradient(tf):
    r = tf.Variable(0.1)
    phi = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=3),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r=r, phi=phi)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [r, phi])

    expected_state_vector = (
        1
        / np.sqrt(np.cosh(r))
        * np.array([1, 0, -np.exp(1j * phi) * np.tanh(r) * np.sqrt(2) / 2])
    )
    expected_jacobian = np.real(
        np.array(
            [
                [
                    -np.sinh(r) / (2 * np.cosh(r) ** (3 / 2)),
                    0,
                    np.exp(1j * phi)
                    * (3 * np.sinh(r) ** 2 - 2 * np.cosh(r) ** 2)
                    / (2 * np.sqrt(2) * np.cosh(r) ** (5 / 2)),
                ],
                [
                    0,
                    0,
                    -1j
                    * np.exp(1j * phi)
                    * np.tanh(r)
                    * np.sqrt(2)
                    / (2 * np.sqrt(np.cosh(r))),
                ],
            ]
        )
    )

    assert np.allclose(state_vector, expected_state_vector)
    assert np.allclose(jacobian, expected_jacobian)


def test_displaced_state_Squeezing_state_vector_gradient(tf):
    r = tf.Variable(0.1)
    phi = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=1,
        config=pq.Config(cutoff=5),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 5)

            pq.Q(0) | pq.Squeezing(r=r, phi=phi)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [r, phi])

    assert np.allclose(
        state_vector,
        [
            0.63519669 + 0.00657421j,
            0.50710742 + 0.37614499j,
            0.11111278 + 0.38482443j,
            -0.06637907 + 0.15485146j,
            -0.04802964 + 0.0180018j,
        ],
    )
    assert np.allclose(
        jacobian,
        [
            [0.27494016, 0.12359607, -0.23514226, 0.12369870, 0.53532110],
            [0.00687977, 0.02229462, 0.05786623, 0.07308305, 0.04782758],
        ],
    )


def test_displaced_state_Beamsplitter_state_vector_gradient(tf):
    theta = tf.Variable(0.05)
    phi = tf.Variable(np.pi / 3)

    simulator = pq.PureFockSimulator(
        d=2,
        config=pq.Config(cutoff=6),
        connector=pq.TensorflowConnector(),
    )

    with tf.GradientTape() as tape:
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 5)
            pq.Q(1) | pq.Displacement(r=1.0, phi=np.pi / 5)

            pq.Q(0, 1) | pq.Beamsplitter(theta, phi)

        state = simulator.execute(program).state

        state_vector = state.state_vector

    jacobian = tape.jacobian(state_vector, [theta, phi])
    correct_state_vector = [
        0.36787944 + 0.0j,
        0.28045205 + 0.22344226j,
        0.29532688 + 0.23424946j,
        0.05521643 + 0.24089812j,
        0.08286355 + 0.3579549j,
        0.06217113 + 0.26594453j,
        -0.0601728 + 0.12539207j,
        -0.1090666 + 0.228548j,
        -0.11413312 + 0.24050364j,
        -0.06895407 + 0.14611768j,
        -0.06101656 + 0.02952233j,
        -0.12814978 + 0.06234706j,
        -0.16481668 + 0.08062809j,
        -0.14131579 + 0.06951122j,
        -0.07419819 + 0.03669684j,
        -0.02882161 - 0.0065087j,
        -0.06778149 - 0.01515268j,
        -0.10081657 - 0.02230856j,
        -0.10603167 - 0.02322176j,
        -0.07885371 - 0.01709068j,
        -0.03708827 - 0.00795441j,
    ]

    correct_jacobian = [
        [
            0.0,
            -0.35052943,
            -0.05328067,
            -0.49699682,
            -0.6256707,
            -0.3798101,
            -0.24836758,
            -0.54263073,
            -0.6643838,
            -0.45959768,
            0.02016149,
            -0.08133037,
            -0.26358867,
            -0.36225897,
            -0.26171276,
            0.10512505,
            0.17760125,
            0.16063823,
            0.06011391,
            -0.03620036,
            -0.05506517,
        ],
        [
            0.0,
            0.00747839,
            -0.01828559,
            -0.00636516,
            -0.01746456,
            -0.01902903,
            -0.01710664,
            -0.02221716,
            -0.01405889,
            -0.00294601,
            -0.01389679,
            -0.01826834,
            -0.00945533,
            0.00392901,
            0.00838149,
            -0.00578762,
            -0.00771638,
            -0.00271062,
            0.006368,
            0.01159076,
            0.00867541,
        ],
    ]

    assert np.allclose(state_vector, correct_state_vector)
    assert np.allclose(jacobian, correct_jacobian)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, List, Callable

from functools import lru_cache

import numpy as np
import numba as nb

from scipy.special import factorial, comb

from piquasso.api.connector import BaseConnector

from ..state import PureFockState
from ...calculations import (
    calculate_interferometer_helper_indices,
    calculate_index_list_for_appling_interferometer,
)
from piquasso.instructions import gates

from piquasso.api.result import Result


def passive_linear(
    state: PureFockState, instruction: gates._PassiveLinearGate, shots: int
) -> Result:
    connector = state._connector

    interferometer: np.ndarray = instruction._get_passive_block(
        state._connector, state._config
    ).astype(state._config.complex_dtype)

    _apply_passive_linear(state, interferometer, instruction.modes, connector)

    return Result(state=state)


def _apply_passive_linear(state, interferometer, modes, connector):
    wrapped = connector.decorator(_do_apply_passive_linear)

    state.state_vector = wrapped(
        state.state_vector,
        interferometer,
        state.d,
        state._config.cutoff,
        modes,
        connector,
    )


def _do_apply_passive_linear(state_vector, interferometer, d, cutoff, modes, connector):
    subspace_transformations = _get_interferometer_on_fock_space(
        interferometer, cutoff, connector
    )

    return _apply_passive_gate_matrix_to_state(
        state_vector, subspace_transformations, d, cutoff, modes, connector
    )


def _get_interferometer_on_fock_space(interferometer, cutoff, connector):
    def _get_interferometer_with_gradient_callback(interferometer):
        interferometer = connector.preprocess_input_for_custom_gradient(interferometer)
        index_tuple = calculate_interferometer_helper_indices(
            d=len(interferometer), cutoff=cutoff
        )

        subspace_representations = connector.calculate_interferometer_on_fock_space(
            interferometer, index_tuple
        )
        grad = _calculate_interferometer_gradient_on_fock_space(
            interferometer,
            connector,
            subspace_representations,
            index_tuple,
        )

        return subspace_representations, grad

    wrapped = connector.custom_gradient(_get_interferometer_with_gradient_callback)

    return wrapped(interferometer)


@nb.njit(cache=True)
def _calculate_subspace_grad(
    row_index,
    col_index,
    previous_subspace_representation,
    subspace_index_tuple,
    interferometer,
    previous_subspace_grad,
):
    (
        subspace_indices,
        first_nonzero_indices,
        first_subspace_indices,
        sqrt_occupation_numbers,
        sqrt_first_occupation_numbers,
    ) = subspace_index_tuple

    matrix_dim = sqrt_occupation_numbers.shape[0]

    subspace_grad = np.zeros(shape=(matrix_dim, matrix_dim), dtype=interferometer.dtype)

    for jdx in range(matrix_dim):
        for idx, first_nonzero_index in enumerate(first_nonzero_indices):
            if first_nonzero_index != row_index:
                continue

            subspace_grad[idx, jdx] += (
                previous_subspace_representation[
                    first_subspace_indices[idx],
                    subspace_indices[jdx, col_index],
                ]
                * sqrt_occupation_numbers[jdx, col_index]
            )

        for idx in range(matrix_dim):
            for kdx in range(sqrt_occupation_numbers.shape[1]):
                subspace_grad[idx, jdx] += (
                    sqrt_occupation_numbers[jdx, kdx]
                    * interferometer[first_nonzero_indices[idx], kdx]
                    * previous_subspace_grad[
                        first_subspace_indices[idx], subspace_indices[jdx, kdx]
                    ]
                )

            subspace_grad[idx, jdx] /= sqrt_first_occupation_numbers[idx]

    return subspace_grad


def _calculate_interferometer_gradient_on_fock_space(
    interferometer, connector, subspace_representations, index_tuple
):
    def interferometer_gradient(*upstream):
        tf = connector._tf
        fallback_np = connector.fallback_np

        static_valued = tf.get_static_value(upstream[0]) is not None

        if static_valued:
            np = connector.fallback_np
            upstream = [x.numpy() for x in upstream]
        else:
            np = connector.np

        d = len(interferometer)
        cutoff = len(index_tuple[0]) + 2

        (
            subspace_index_tensor,
            first_nonzero_index_tensor,
            first_subspace_index_tensor,
            sqrt_occupation_numbers_tensor,
            sqrt_first_occupation_numbers_tensor,
        ) = index_tuple

        full_kl_grad = []
        for row_index in range(d):
            full_kl_grad.append([0] * d)
            for col_index in range(d):
                second_subspace_grad = fallback_np.zeros(
                    shape=interferometer.shape, dtype=interferometer.dtype
                )
                second_subspace_grad[row_index, col_index] = 1.0

                previous_subspace_grad = second_subspace_grad

                for p in range(2, cutoff):
                    previous_subspace_representation = subspace_representations[p - 1]

                    subspace_index_tuple = (
                        subspace_index_tensor[p - 2],
                        first_nonzero_index_tensor[p - 2],
                        first_subspace_index_tensor[p - 2],
                        sqrt_occupation_numbers_tensor[p - 2],
                        sqrt_first_occupation_numbers_tensor[p - 2],
                    )

                    subspace_grad = _calculate_subspace_grad(
                        row_index,
                        col_index,
                        previous_subspace_representation,
                        subspace_index_tuple,
                        interferometer,
                        previous_subspace_grad,
                    )

                    full_kl_grad[row_index][col_index] += np.einsum(
                        "ij,ij", upstream[p], fallback_np.conj(subspace_grad)
                    )

                    previous_subspace_grad = subspace_grad

        partial_result = connector.np.array(full_kl_grad, dtype=interferometer.dtype)
        one_particle_term = upstream[1]
        result = partial_result + one_particle_term

        return result

    return interferometer_gradient


def _apply_passive_gate_matrix_to_state(
    state_vector: np.ndarray,
    subspace_transformations: List[np.ndarray],
    d: int,
    cutoff: int,
    modes: Tuple[int, ...],
    connector: BaseConnector,
) -> None:
    def _apply_interferometer_matrix(state_vector, subspace_transformations):
        state_vector = connector.preprocess_input_for_custom_gradient(state_vector)

        subspace_transformations = [
            connector.preprocess_input_for_custom_gradient(matrix)
            for matrix in subspace_transformations
        ]

        index_list = calculate_index_list_for_appling_interferometer(
            modes,
            d,
            cutoff,
        )

        new_state_vector = _calculate_state_vector_after_interferometer(
            state_vector,
            subspace_transformations,
            index_list,
            connector,
        )

        grad = _create_linear_passive_gate_gradient_function(
            state_vector,
            subspace_transformations,
            index_list,
            connector,
        )
        return new_state_vector, grad

    wrapped = connector.custom_gradient(_apply_interferometer_matrix)

    return wrapped(state_vector, subspace_transformations)


def _calculate_state_vector_after_interferometer(
    state_vector: np.ndarray,
    subspace_transformations: List[np.ndarray],
    index_list: List[np.ndarray],
    connector: BaseConnector,
) -> np.ndarray:
    np = connector.forward_pass_np

    new_state_vector = np.empty_like(state_vector)

    is_batch = len(state_vector.shape) == 2

    einsum_string = "ij,jkl->ikl" if is_batch else "ij,jk->ik"

    for n, indices in enumerate(index_list):
        new_state_vector = connector.assign(
            new_state_vector,
            indices,
            np.einsum(
                einsum_string, subspace_transformations[n], state_vector[indices]
            ),
        )

    return new_state_vector


def _create_linear_passive_gate_gradient_function(
    state_vector: np.ndarray,
    subspace_transformations: List[np.ndarray],
    index_list: List[np.ndarray],
    connector: BaseConnector,
) -> Callable:
    def applying_interferometer_gradient(upstream):
        tf = connector._tf
        fallback_np = connector.fallback_np

        static_valued = tf.get_static_value(upstream) is not None

        if static_valued:
            np = connector.fallback_np
            upstream = upstream.numpy()
        else:
            np = connector.np

        is_batch = len(state_vector.shape) == 2

        matrix_einsum_string = "ijl,kjl->ki" if is_batch else "ij,kj->ki"
        initial_state_einsum_string = "ji,jkl->ikl" if is_batch else "ji,jk->ik"

        reshape_arg = (-1, state_vector.shape[1]) if is_batch else (-1,)

        unordered_gradient_by_initial_state = []
        order_by = []

        gradient_by_matrix = []

        conjugated_state_vector = np.conj(state_vector)

        for n, indices in enumerate(index_list):
            matrix = np.conj(subspace_transformations[n])
            sliced_upstream = upstream[indices]
            state_vector_slice = conjugated_state_vector[indices]

            order_by.append(indices.reshape(-1))
            product = np.einsum(initial_state_einsum_string, matrix, sliced_upstream)
            unordered_gradient_by_initial_state.append(product.reshape(*reshape_arg))

            gradient_by_matrix.append(
                np.einsum(matrix_einsum_string, state_vector_slice, sliced_upstream)
            )

        gradient_by_initial_state = np.concatenate(unordered_gradient_by_initial_state)[
            fallback_np.concatenate(order_by).argsort()
        ]

        if static_valued:
            gradient_by_initial_state = tf.constant(gradient_by_initial_state)
            gradient_by_matrix = [tf.constant(matrix) for matrix in gradient_by_matrix]

        return gradient_by_initial_state, gradient_by_matrix

    return applying_interferometer_gradient


@lru_cache(maxsize=None)
def _beamsplitter5050_coeff(n, m, N):
    return (
        2 ** (-(n + m) / 2)
        * np.sqrt(factorial(N) * factorial(n + m - N) / (factorial(n) * factorial(m)))
        * (-1) ** n
        * np.sum([comb(n, j) * comb(m, N - j) * (-1) ** j for j in range(N + 1)])
    )


@lru_cache(maxsize=None)
def _calculate_beamsplitter5050_for_symmetric_subspace(number_of_particles):
    """
    NOTE: Further optimizations may be done here.
    """
    dimension = number_of_particles + 1

    matrix = np.empty(shape=(dimension, dimension))

    for row_index, n in enumerate(range(number_of_particles, -1, -1)):
        m = number_of_particles - n
        for col_index, N in enumerate(range(number_of_particles, -1, -1)):
            matrix[row_index, col_index] = _beamsplitter5050_coeff(n, m, N)

    return matrix


@lru_cache(maxsize=None)
def _calculate_beamsplitter5050_on_fock_space(cutoff):
    return [
        _calculate_beamsplitter5050_for_symmetric_subspace(number_of_particles)
        for number_of_particles in range(cutoff)
    ]


def beamsplitter5050(
    state: PureFockState, instruction: gates._PassiveLinearGate, shots: int
) -> Result:
    _apply_beamsplitter5050(state, instruction.modes)

    return Result(state=state)


def _apply_beamsplitter5050(state, modes):
    cutoff = state._config.cutoff
    subspace_transformations = _calculate_beamsplitter5050_on_fock_space(cutoff)

    state.state_vector = _apply_passive_gate_matrix_to_state(
        state_vector=state.state_vector,
        subspace_transformations=subspace_transformations,
        d=state.d,
        cutoff=cutoff,
        modes=modes,
        connector=state._connector,
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_beamsplitter_with_theta_pi_over_4():
    with pq.Program() as program:
        pq.Q(1) | pq.StateVector([1])

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0, 0.5, 0.5, 0, 0, 0],
    )


def test_beamsplitter5050_equivalence():
    with pq.Program() as preparation:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 5)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7, phi=np.pi / 11)

    with pq.Program() as program_with_parametrized_beamsplitter:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=0.0)

    with pq.Program() as program_with_50_50_beamsplitter:
        pq.Q() | preparation

        pq.Q(0, 1) | pq.Beamsplitter5050()

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    state_with_parametrized_beamsplitter = simulator.execute(
        program_with_parametrized_beamsplitter
    ).state

    state_with_50_50_beamsplitter = simulator.execute(
        program_with_50_50_beamsplitter
    ).state

    assert state_with_50_50_beamsplitter == state_with_parametrized_beamsplitter


def test_beamsplitter():
    with pq.Program() as program:
        pq.Q(1) | 1 * pq.StateVector([1])

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.3454915, 0.6545085, 0.0, 0.0, 0.0],
    )


def test_beamsplitter_multiple_particles():
    with pq.Program() as program:
        pq.Q(1) | pq.StateVector([1]) / 2
        pq.Q(1) | pq.StateVector([2]) / 2
        pq.Q(0) | pq.StateVector([2]) / np.sqrt(2)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.08637288, 0.16362712, 0.32397979, 0.17929466, 0.24672554],
    )


def test_beamsplitter_leaves_vacuum_unchanged():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0]) / 2
        pq.Q() | pq.StateVector([0, 1]) / np.sqrt(2)
        pq.Q() | pq.StateVector([0, 2]) / 2

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.25, 0.17274575, 0.32725425, 0.02984109, 0.11306356, 0.10709534],
    )


def test_multiple_beamsplitters():
    with pq.Program() as program:
        pq.Q(2) | pq.StateVector([1]) * 1

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=1.5 * np.pi)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    )


def test_multiple_beamsplitters_with_multiple_particles():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 1]) / 2
        pq.Q() | pq.StateVector([0, 0, 2]) / 2
        pq.Q() | pq.StateVector([0, 1, 1]) / np.sqrt(2)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=1.5 * np.pi)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.0625, 0.1875, 0.0, 0.0625, 0.1875, 0.109375, 0.15625, 0.234375],
    )


def test_phaseshift():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2
        pq.Q() | pq.StateVector([0, 2]) / np.sqrt(2)
        pq.Q() | pq.StateVector([1, 1]) / 2

        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.0, 0.25, 0.5],
    )


def test_fourier():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2
        pq.Q() | pq.StateVector([0, 2]) / np.sqrt(2)
        pq.Q() | pq.StateVector([1, 1]) / 2

        pq.Q(0) | pq.Fourier()

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.25, 0.0, 0.25, 0.5],
    )


def test_mach_zehnder():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1]) / 2
        pq.Q() | pq.StateVector([0, 2]) / np.sqrt(2)
        pq.Q() | pq.StateVector([1, 1]) / 2

        pq.Q(0, 1) | pq.MachZehnder(int_=np.pi / 3, ext=np.pi / 4)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.1875, 0.0625, 0.60463966, 0.09690689, 0.04845345],
    )


def test_beamsplitters_and_phaseshifters_with_multiple_particles():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 1]) / 2
        pq.Q() | pq.StateVector([0, 0, 2]) / 2
        pq.Q() | pq.StateVector([0, 1, 1]) / np.sqrt(2)

        pq.Q(0) | pq.Phaseshifter(phi=np.pi / 3)
        pq.Q(1) | pq.Phaseshifter(phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4, phi=4 * np.pi / 5)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=3 * np.pi / 2)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [
            0.0,
            0.0,
            0.0625,
            0.1875,
            0.0,
            0.0625,
            0.1875,
            0.04308374,
            0.02366748,
            0.43324878,
        ],
    )


def test_interferometer():
    T = np.array(
        [
            [0.5, 0.53033009 + 0.53033009j, 0.21650635 + 0.375j],
            [-0.61237244 + 0.61237244j, 0.4330127, 0.24148146 + 0.06470476j],
            [0, -0.48296291 + 0.12940952j, 0.8660254],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 1]) / 2
        pq.Q() | pq.StateVector([0, 0, 2]) / 2
        pq.Q() | pq.StateVector([0, 1, 1]) / np.sqrt(2)

        pq.Q(0, 1, 2) | pq.Interferometer(matrix=T)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(sum(state.fock_probabilities), 1)
    assert np.allclose(
        state.fock_probabilities,
        [
            0.0,
            0.046875,
            0.015625,
            0.1875,
            0.17307537,
            0.11538358,
            0.32090931,
            0.0192306,
            0.10696977,
            0.01443139,
        ],
    )


def test_kerr():
    xi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 2, 1])

        pq.Q(1) | pq.Kerr(xi=xi)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 1

    assert np.isclose(nonzero_elements[0][0], -np.exp(1j * np.pi / 3))
    assert nonzero_elements[0][1] == (0, 2, 1)


def test_cross_kerr():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 2, 1])

        pq.Q(1, 2) | pq.CrossKerr(xi=np.pi / 2)

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 1

    assert np.isclose(nonzero_elements[0][0], -1)
    assert nonzero_elements[0][1] == (0, 2, 1)


def test_cubic_phase():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.CubicPhase(gamma=0.1)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state

    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 5.0



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np

from scipy.linalg import polar, coshm, sinhm, block_diag
from scipy.special import factorial2

import piquasso as pq
from piquasso.api.exceptions import InvalidParameter


def test_xxpp_representation(state, assets):
    assert np.allclose(
        assets.load("expected_xxpp_mean"),
        state.xxpp_mean_vector,
    )
    assert np.allclose(
        assets.load("expected_xxpp_correlation"),
        state.xxpp_correlation_matrix,
    )
    assert np.allclose(
        assets.load("expected_xxpp_covariance"),
        state.xxpp_covariance_matrix,
    )


def test_xpxp_representation(state, assets):
    assert np.allclose(
        assets.load("expected_xpxp_mean"),
        state.xpxp_mean_vector,
    )
    assert np.allclose(
        assets.load("expected_xpxp_correlation"),
        state.xpxp_correlation_matrix,
    )
    assert np.allclose(
        assets.load("expected_xpxp_covariance"),
        state.xpxp_covariance_matrix,
    )


def test_set_xpxp_mean_vector_invalid_shape_raises_InvalidState(state):
    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.xpxp_mean_vector = np.array([1, 2, 3])

    assert "Invalid 'mean' vector shape" in error.value.args[0]


def test_set_xpxp_mean_vector_invalid_shape_validate_False():
    state = pq.GaussianState(
        2, connector=pq.NumpyConnector(), config=pq.Config(validate=False)
    )

    state.xpxp_mean_vector = np.array([1, 2, 3])


def test_set_xpxp_covariance_matrix_invalid_shape_raises_InvalidState(state):
    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.xpxp_covariance_matrix = np.array([1, 2, 3])

    assert "Invalid 'cov' matrix shape" in error.value.args[0]


def test_set_xpxp_covariance_matrix_invalid_shape_validate_False():
    state = pq.GaussianState(
        2, connector=pq.NumpyConnector(), config=pq.Config(validate=False)
    )

    with pytest.raises(Exception) as exc:
        state.xpxp_covariance_matrix = np.array([1, 2, 3])

    assert not isinstance(exc, pq.api.exceptions.PiquassoException)


def test_representation_roundtrip(state):
    initial_mean_vector = state.xpxp_mean_vector
    initial_covariance_matrix = state.xpxp_covariance_matrix

    state.xpxp_mean_vector = initial_mean_vector
    state.xpxp_covariance_matrix = initial_covariance_matrix

    final_mean_vector = state.xpxp_mean_vector
    final_covariance_matrix = state.xpxp_covariance_matrix

    assert np.allclose(final_mean_vector, initial_mean_vector)
    assert np.allclose(final_covariance_matrix, initial_covariance_matrix)


def test_representation_roundtrip_at_different_HBAR(state):
    state._config.hbar = 42

    initial_mean_vector = state.xpxp_mean_vector
    initial_covariance_matrix = state.xpxp_covariance_matrix

    state.xpxp_mean_vector = initial_mean_vector
    state.xpxp_covariance_matrix = initial_covariance_matrix

    final_mean_vector = state.xpxp_mean_vector
    final_covariance_matrix = state.xpxp_covariance_matrix

    assert np.allclose(final_mean_vector, initial_mean_vector)
    assert np.allclose(final_covariance_matrix, initial_covariance_matrix)


def test_xp_representation_roundtrip(state):
    initial_mean = state.xxpp_mean_vector
    initial_cov = state.xxpp_covariance_matrix

    state.xxpp_mean_vector = initial_mean
    state.xxpp_covariance_matrix = initial_cov

    final_mean = state.xxpp_mean_vector
    final_cov = state.xxpp_covariance_matrix

    assert np.allclose(final_mean, initial_mean)
    assert np.allclose(final_cov, initial_cov)


def test_xp_representation_roundtrip_at_different_HBAR(state):
    state._config.hbar = 42

    initial_mean = state.xxpp_mean_vector
    initial_cov = state.xxpp_covariance_matrix

    state.xxpp_mean_vector = initial_mean
    state.xxpp_covariance_matrix = initial_cov

    final_mean = state.xxpp_mean_vector
    final_cov = state.xxpp_covariance_matrix

    assert np.allclose(final_mean, initial_mean)
    assert np.allclose(final_cov, initial_cov)


def test_wigner_function(state, assets):
    actual_result = state.wigner_function(
        positions=[[1.0, 3.0, 5.0], [5.0, 7.0, 9.0]],
        momentums=[[2.0, 4.0, 6.0], [6.0, 8.0, 0.0]],
    )

    expected_result = assets.load("expected_wigner_function_result")
    assert np.allclose(expected_result, actual_result)


def test_wigner_plot_1D():
    # Create a Gaussian state
    config = pq.Config(cutoff=10, hbar=42)
    dim = 1
    with pq.Program() as initialization:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.10, phi=np.angle(1 - 0.5j))
        pq.Q(0) | pq.Squeezing(r=0.10)
    simulator = pq.GaussianSimulator(d=dim, config=config)
    state = simulator.execute(initialization).state
    state.validate()

    x = np.linspace(-5, 5, 20)
    p = np.linspace(-5, 5, 20)

    x = np.array([x[i : i + dim] for i in range(0, len(x), dim)])
    p = np.array([p[i : i + dim] for i in range(0, len(p), dim)])
    mode = 0
    try:
        state.plot_wigner(positions=x, momentums=p, mode=mode)
    except Exception as e:
        pytest.fail(f"Plotting failed with exception: {e}")


def test_GaussState_plot_wigner_function_raises_InvalidModes_for_multiple_modes():
    alpha = 1 - 0.5j
    config = pq.Config(cutoff=10, hbar=42)
    dim = 1
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = pq.GaussianSimulator(d=dim, config=config)
    state = simulator.execute(program).state
    state.validate()

    with pytest.raises(pq.api.exceptions.InvalidModes):
        x = np.linspace(-5, 5, 20)
        p = np.linspace(-5, 5, 20)
        mode = 2
        state.plot_wigner(x, p, mode=mode)


def test_GaussState_plot_wigner_function_raises_InvalidModes_for_multiple_dimensions():
    alpha = 1 - 0.5j
    config = pq.Config(cutoff=10, hbar=42)
    dim = 2
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = pq.GaussianSimulator(d=dim, config=config)
    state = simulator.execute(program).state
    state.validate()

    with pytest.raises(pq.api.exceptions.InvalidModes):
        x = np.linspace(-5, 5, 20)
        p = np.linspace(-5, 5, 20)
        state.plot_wigner(x, p)


def test_reduced_rotated_mean_and_covariance(state, assets):
    modes = (0, 2)
    phi = np.pi / 2

    mean_vector, covariance_matrix = state.xpxp_reduced_rotated_mean_and_covariance(
        modes, phi
    )

    expected_mean = assets.load("expected_mean")
    expected_cov = assets.load("expected_covariance")

    assert np.allclose(mean_vector, expected_mean)
    assert np.allclose(covariance_matrix, expected_cov)


def test_rotated():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1.0)
        pq.Q(1) | pq.Displacement(r=2.0)
        pq.Q(2) | pq.Displacement(r=3.0, phi=np.pi / 2)

        pq.Q(0, 1) | pq.Squeezing2(r=np.log(2.0), phi=0.0)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=0)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state
    state.validate()

    phi = np.pi / 2
    rotated_state = state.rotated(phi)

    expected_rotated_mean_vector = np.array([0.0, 6.5, 0.0, -5.5, 6.0, 0.0])
    expected_rotated_covariance_matrix = np.array(
        [
            [4.25, 0.0, 3.75, 0.0, 0.0, 0.0],
            [0.0, 4.25, 0.0, -3.75, 0.0, 0.0],
            [3.75, 0.0, 4.25, -0.0, 0.0, 0.0],
            [0.0, -3.75, -0.0, 4.25, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 2.0],
        ]
    )

    assert np.allclose(
        expected_rotated_mean_vector,
        rotated_state.xpxp_mean_vector,
    )
    assert np.allclose(
        expected_rotated_covariance_matrix,
        rotated_state.xpxp_covariance_matrix,
    )


def test_rotated_state_inherits_config():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1.0)
        pq.Q(1) | pq.Displacement(r=2.0)
        pq.Q(2) | pq.Displacement(r=3.0, phi=np.pi / 2)

        pq.Q(0, 1) | pq.Squeezing2(r=np.log(2.0), phi=0.0)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=0)

    config = pq.Config(hbar=1, seed_sequence=42)

    simulator = pq.GaussianSimulator(d=3, config=config)
    state = simulator.execute(program).state
    state.validate()

    reduced_state = state.rotated(phi=np.pi / 5)

    assert state._config == config
    assert reduced_state._config == state._config


def test_reduced():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1.0)
        pq.Q(1) | pq.Displacement(r=2.0)
        pq.Q(2) | pq.Displacement(r=3.0, phi=np.pi / 2)

        pq.Q(0, 1) | pq.Squeezing2(r=np.log(2.0), phi=0.0)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=0)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state
    state.validate()

    modes = (0, 2)

    reduced_state = state.reduced(modes)

    expected_reduced_mean_vector = np.array([-6.5, 0.0, 0.0, 6.0])
    expected_reduced_covariance_matrix = np.array(
        [
            [4.25, 0.0, 0.0, 0.0],
            [0.0, 4.25, 0.0, 0.0],
            [0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 0.0, 2.0],
        ]
    )

    assert np.allclose(
        expected_reduced_mean_vector,
        reduced_state.xpxp_mean_vector,
    )
    assert np.allclose(
        expected_reduced_covariance_matrix,
        reduced_state.xpxp_covariance_matrix,
    )


def test_reduced_state_inherits_config():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1.0)
        pq.Q(1) | pq.Displacement(r=2.0)
        pq.Q(2) | pq.Displacement(r=3.0, phi=np.pi / 2)

        pq.Q(0, 1) | pq.Squeezing2(r=np.log(2.0), phi=0.0)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=0)

    config = pq.Config(hbar=1, seed_sequence=42)

    simulator = pq.GaussianSimulator(d=3, config=config)
    state = simulator.execute(program).state
    state.validate()

    reduced_state = state.reduced(modes=(0,))

    assert state._config == config
    assert reduced_state._config == state._config


def test_GaussianState_get_marginal_fock_probabilities():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=1.0)
        pq.Q(1) | pq.Displacement(r=2.0)
        pq.Q(2) | pq.Displacement(r=3.0, phi=np.pi / 2)

        pq.Q(0, 1) | pq.Squeezing2(r=np.log(2.0), phi=0.0)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 2, phi=0)

    simulator = pq.GaussianSimulator(d=3, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state
    state.validate()

    modes = (0, 2)
    expected_probabilities = state.reduced(modes).fock_probabilities
    actual_probabilities = state.get_marginal_fock_probabilities(modes)

    assert np.allclose(actual_probabilities, expected_probabilities)


def test_vacuum_covariance_is_proportional_to_identity():
    d = 2

    simulator = pq.GaussianSimulator(d=d)

    state = simulator.create_initial_state()

    expected_xpxp_mean = np.zeros(2 * d)
    expected_xpxp_covariance_matrix = np.identity(2 * d) * state._config.hbar

    assert np.allclose(state.xpxp_mean_vector, expected_xpxp_mean)
    assert np.allclose(state.xpxp_covariance_matrix, expected_xpxp_covariance_matrix)


def test_mean_and_covariance(state, assets):
    expected_mean_vector = assets.load("expected_mean") * np.sqrt(
        2 * state._config.hbar
    )

    expected_covariance_matrix = assets.load("expected_cov") * state._config.hbar

    assert np.allclose(state.xpxp_mean_vector, expected_mean_vector)
    assert np.allclose(state.xpxp_covariance_matrix, expected_covariance_matrix)


def test_mean_and_covariance_with_different_HBAR(state, assets):
    state._config.hbar = 42

    expected_mean_vector = assets.load("expected_mean") * np.sqrt(
        2 * state._config.hbar
    )
    expected_covariance_matrix = assets.load("expected_cov") * state._config.hbar

    assert np.allclose(state.xpxp_mean_vector, expected_mean_vector)
    assert np.allclose(state.xpxp_covariance_matrix, expected_covariance_matrix)


def test_mean_is_scaled_with_squared_HBAR(state, assets):
    scaling = 14

    mean_default_hbar = state.xpxp_mean_vector

    state._config.hbar *= scaling

    mean_different_hbar = state.xpxp_mean_vector

    assert np.allclose(mean_default_hbar * np.sqrt(scaling), mean_different_hbar)


def test_covariance_is_scaled_with_HBAR(state):
    scaling = 14

    cov_default_hbar = state.xpxp_covariance_matrix

    state._config.hbar *= scaling

    cov_different_hbar = state.xpxp_covariance_matrix

    assert np.allclose(cov_default_hbar * scaling, cov_different_hbar)


def test_complex_covariance(state, assets):
    expected_complex_covariance = assets.load("expected_complex_covariance")

    assert np.allclose(state.complex_covariance, expected_complex_covariance)


def test_complex_covariance_does_not_scale_with_HBAR(state):
    complex_covariance_default_hbar = state.complex_covariance

    state._config.hbar = 42

    complex_covariance_different_hbar = state.complex_covariance

    assert np.allclose(
        complex_covariance_default_hbar,
        complex_covariance_different_hbar,
    )


def test_complex_covariance_transformed_directly_from_xxpp_cov(state):
    xxpp_cov = state.xxpp_covariance_matrix
    d = len(xxpp_cov) // 2

    W = (1 / np.sqrt(2)) * np.block(
        [
            [np.identity(d), 1j * np.identity(d)],
            [np.identity(d), -1j * np.identity(d)],
        ]
    )

    complex_covariance = W @ xxpp_cov @ W.conj().T / state._config.hbar

    assert np.allclose(complex_covariance, state.complex_covariance)


def test_complex_displacement(state, assets):
    expected_complex_displacement = assets.load("expected_complex_displacement")

    assert np.allclose(
        state.complex_displacement,
        expected_complex_displacement,
    )


def test_complex_displacement_do_not_scale_with_HBAR(state):
    complex_displacement_default_hbar = state.complex_displacement

    state._config.hbar = 42

    complex_displacement_different_hbar = state.complex_displacement

    assert np.allclose(
        complex_displacement_default_hbar,
        complex_displacement_different_hbar,
    )


def test_quadratic_expectation_with_nonsymmetric_quadratic_coefficients(state):
    nonsymmetric_quadratic_coefficients = np.array(
        [
            [0, 1, 1, 1, 1, 1],
            [0, 0, 1, 1, 1, 1],
            [0, 0, 0, 1, 1, 1],
            [0, 0, 0, 0, 1, 1],
            [0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0],
        ]
    )

    any_vector = np.empty(2 * state.d)

    with pytest.raises(InvalidParameter):
        state.quadratic_polynomial_expectation(
            A=nonsymmetric_quadratic_coefficients,
            b=any_vector,
        )


def test_quadratic_expectation(state):
    rotation_angle = np.pi / 3
    constant_term = 100
    quadratic_coefficients = np.array(
        [
            [0, 1, 2, 0, 2, 0],
            [1, 0, 0, 1, 0, 1],
            [2, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [2, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
        ]
    )
    linear_coefficients = np.ones(2 * state.d)

    expectation_value = state.quadratic_polynomial_expectation(
        A=quadratic_coefficients,
        b=linear_coefficients,
        c=constant_term,
        phi=rotation_angle,
    )

    assert np.isclose(
        expectation_value,
        97.78556059428445,
    )


def test_mean_photon_number_vaccum():
    with pq.Program() as program:
        pass

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    assert np.isclose(0.0, state.mean_photon_number((0, 1, 2)))
    assert np.isclose(0.0, state.mean_photon_number((0, 1)))
    assert np.isclose(0.0, state.mean_photon_number((0,)))


def test_mean_photon_number():
    alpha = 0.5 + 1j
    r = 1.0
    phi = 3.0
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(1) | pq.Squeezing(r=r, phi=phi)
        pq.Q(2) | pq.Squeezing(r=r, phi=phi)
        pq.Q(2) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    mean_photon_number_first_mode = np.abs(alpha) ** 2
    mean_photon_number_second_mode = np.sinh(r) ** 2
    total_mean_photon_number = 2 * (np.abs(alpha) ** 2 + np.sinh(r) ** 2)

    assert np.isclose(mean_photon_number_first_mode, state.mean_photon_number((0,)))
    assert np.isclose(mean_photon_number_second_mode, state.mean_photon_number((1,)))
    assert np.isclose(total_mean_photon_number, state.mean_photon_number((0, 1, 2)))


def test_variance_photon_number():
    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=0.2, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.8, phi=np.pi / 2)
        pq.Q(2) | pq.Displacement(r=1.0)

    simulator = pq.GaussianSimulator(d=4)
    state = simulator.execute(program).state

    assert np.isclose(state.variance_photon_number((0,)), 0.04)
    assert np.isclose(
        state.variance_photon_number((1,)), 2 * np.sinh(0.8) ** 2 * np.cosh(0.8) ** 2
    )
    assert np.isclose(state.variance_photon_number((2,)), 1.0)
    assert np.isclose(state.variance_photon_number((3,)), 0.0)


def test_displaced_state_variance_photon_number_multimode():
    r_1 = 0.2
    r_2 = 0.3

    with pq.Program() as program:
        pq.Q(0) | pq.Displacement(r=r_1, phi=np.pi / 3)
        pq.Q(1) | pq.Displacement(r=r_2, phi=np.pi / 7)
        pq.Q(0, 1) | pq.Beamsplitter5050()

    simulator = pq.GaussianSimulator(d=4)
    state = simulator.execute(program).state

    assert np.isclose(state.variance_photon_number(), r_1**2 + r_2**2)


def test_GaussianState_get_particle_detection_probability():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    probability = state.get_particle_detection_probability(occupation_number=(0, 2))

    assert np.isclose(probability, 0.0012355308401079989)


def test_GaussianState_threshold_and_particle_resolved_vacuum_detection_equivalence():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    vacuum_probability_1 = state.get_threshold_detection_probability(
        occupation_number=(0, 0)
    )
    vacuum_probability_2 = state.get_particle_detection_probability(
        occupation_number=(0, 0)
    )

    assert np.isclose(vacuum_probability_1, vacuum_probability_2)


def test_displaced_GaussianState_threshold_and_particle_resolved_vacuum_detection_equivalence():  # noqa: E501
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 7)
        pq.Q(1) | pq.Displacement(r=0.2, phi=-np.pi / 11)

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 7)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    vacuum_probability_1 = state.get_threshold_detection_probability(
        occupation_number=(0, 0)
    )
    vacuum_probability_2 = state.get_particle_detection_probability(
        occupation_number=(0, 0)
    )

    assert np.isclose(vacuum_probability_1, vacuum_probability_2)


def test_GaussianState_get_threshold_detection_probability_short_occupation_number():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.PiquassoException) as error:
        state.get_threshold_detection_probability(occupation_number=(1,))

    error_message = error.value.args[0]

    assert error_message == (
        "The specified occupation number should have length '2': "
        "occupation_number='(1,)'."
    )


def test_GaussianState_get_threshold_detection_probability_invalid_occupation_number():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.PiquassoException) as error:
        state.get_threshold_detection_probability(occupation_number=(1, 2))

    error_message = error.value.args[0]

    assert error_message == (
        "The specified occupation numbers must only contain '0' or '1': "
        "occupation_number='(1, 2)'."
    )


def test_GaussianState_get_threshold_detection_probability_2_mode():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    probability = state.get_threshold_detection_probability(occupation_number=(1, 0))

    assert np.isclose(probability, 0.0021696454384600313)


def test_displaced_GaussianState_get_threshold_detection_probability_2_mode():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 7)
        pq.Q(1) | pq.Displacement(r=0.2, phi=-np.pi / 11)

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    probability = state.get_threshold_detection_probability(occupation_number=(1, 0))

    assert np.isclose(probability, 0.012210371851308395)


def test_GaussianState_get_threshold_detection_probability_3_mode():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    probability = state.get_threshold_detection_probability(occupation_number=(1, 0, 1))

    assert np.isclose(probability, 0.00093484938797056)


def test_displaced_GaussianState_get_threshold_detection_probability_3_mode():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 7)
        pq.Q(1) | pq.Displacement(r=0.2, phi=-np.pi / 11)

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)
        pq.Q(1) | pq.Squeezing(r=0.05)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    probability = state.get_threshold_detection_probability(occupation_number=(1, 0, 1))

    assert np.isclose(probability, 0.0005136255432135231)


def test_GaussianState_get_threshold_detection_probability_sums_to_one_on_2_mode():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    probability00 = state.get_threshold_detection_probability(occupation_number=(0, 0))
    probability01 = state.get_threshold_detection_probability(occupation_number=(0, 1))
    probability10 = state.get_threshold_detection_probability(occupation_number=(1, 0))
    probability11 = state.get_threshold_detection_probability(occupation_number=(1, 1))

    assert np.isclose(
        probability00 + probability01 + probability10 + probability11, 1.0
    )


def test_displaced_GaussianState_get_threshold_detection_probability_sums_to_one_on_2_mode():  # noqa: E501
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 7)
        pq.Q(1) | pq.Displacement(r=0.2, phi=-np.pi / 11)

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.GaussianSimulator(d=2)
    state = simulator.execute(program).state

    probability00 = state.get_threshold_detection_probability(occupation_number=(0, 0))
    probability01 = state.get_threshold_detection_probability(occupation_number=(0, 1))
    probability10 = state.get_threshold_detection_probability(occupation_number=(1, 0))
    probability11 = state.get_threshold_detection_probability(occupation_number=(1, 1))

    assert np.isclose(
        probability00 + probability01 + probability10 + probability11, 1.0
    )


@pytest.mark.monkey
def test_GaussianState_get_threshold_detection_probability_sums_to_one_random(
    generate_complex_symmetric_matrix, generate_unitary_matrix
):
    d = 3
    squeezing_matrix = generate_complex_symmetric_matrix(d)
    U, r = polar(squeezing_matrix)

    global_phase = generate_unitary_matrix(d)
    passive = global_phase @ coshm(r)
    active = global_phase @ sinhm(r) @ U.conj()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.GaussianTransform(passive=passive, active=active)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    binary_strings = []

    for i in range(2**d, 2 ** (d + 1)):
        binary_strings.append(tuple(int(x) for x in list(bin(i)[3:])))

    probabilities = []

    for binary_string in binary_strings:
        probabilities.append(
            state.get_threshold_detection_probability(occupation_number=binary_string)
        )

    assert np.isclose(sum(probabilities), 1.0)


@pytest.mark.monkey
def test_displaced_GaussianState_get_threshold_detection_probability_sums_to_one_random(
    generate_complex_symmetric_matrix, generate_unitary_matrix
):
    d = 3
    squeezing_matrix = generate_complex_symmetric_matrix(d)
    U, r = polar(squeezing_matrix)

    global_phase = generate_unitary_matrix(d)
    passive = global_phase @ coshm(r)
    active = global_phase @ sinhm(r) @ U.conj()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        for i in range(d):
            pq.Q(i) | pq.Displacement(
                r=np.random.rand(), phi=2 * np.pi * np.random.rand()
            )

        pq.Q() | pq.GaussianTransform(passive=passive, active=active)

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state

    binary_strings = []

    for i in range(2**d, 2 ** (d + 1)):
        binary_strings.append(tuple(int(x) for x in list(bin(i)[3:])))

    probabilities = []

    for binary_string in binary_strings:
        probabilities.append(
            state.get_threshold_detection_probability(occupation_number=binary_string)
        )

    assert np.isclose(sum(probabilities), 1.0)


def test_GaussianState_fidelity():
    with pq.Program() as program_1:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=2, phi=np.pi / 3)
        pq.Q(0) | pq.Displacement(r=0.1, phi=0)

    simulator = pq.GaussianSimulator(d=1)
    state_1 = simulator.execute(program_1).state

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=2, phi=-np.pi / 3)
        pq.Q(0) | pq.Displacement(r=-0.1, phi=0)

    state_2 = simulator.execute(program_2).state

    assert np.isclose(state_2.fidelity(state_1), 0.042150949)
    assert np.isclose(state_2.fidelity(state_1), state_1.fidelity(state_2))
    assert np.isclose(state_2.fidelity(state_2), 1.0)


def test_GaussianState_Vacuum_is_pure():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    assert state.is_pure()


def test_GaussianState_Vacuum_get_purity():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    assert np.isclose(state.get_purity(), 1.0)


def test_GaussianState_purify_vacuum():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.GaussianSimulator(d=3)
    state = simulator.execute(program).state

    purification = state.purify()

    assert state.is_pure()
    assert purification.is_pure()
    assert purification.d == 2 * state.d
    assert purification.reduced(modes=(0, 1, 2)) == state


def test_GaussianState_purify_on_1_mode():
    with pq.Program() as program:
        pq.Q() | pq.Thermal([0.5])

        pq.Q(0) | pq.Squeezing(r=2, phi=np.pi / 3)
        pq.Q(0) | pq.Displacement(r=0.1, phi=0)

    simulator = pq.GaussianSimulator(d=1)
    mixed_state = simulator.execute(program).state

    purification = mixed_state.purify()

    assert not mixed_state.is_pure()
    assert purification.is_pure()
    assert purification.d == 2 * mixed_state.d
    assert purification.reduced(modes=(0,)) == mixed_state


def test_GaussianState_purify_on_2_modes():
    with pq.Program() as program:
        pq.Q() | pq.Thermal([0.5, 1.5])

        pq.Q(0) | pq.Squeezing(r=2, phi=np.pi / 3)
        pq.Q(0) | pq.Displacement(r=0.1, phi=0)
        pq.Q(1) | pq.Displacement(r=0.2, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.GaussianSimulator(d=2)
    mixed_state = simulator.execute(program).state

    purification = mixed_state.purify()

    assert not mixed_state.is_pure()
    assert purification.is_pure()
    assert purification.d == 2 * mixed_state.d
    assert purification.reduced(modes=(0, 1)) == mixed_state


def test_GaussianState_purify_on_3_modes():
    with pq.Program() as program:
        pq.Q() | pq.Thermal([0.5, 1.5, 0.0])

        pq.Q(0) | pq.Squeezing(r=2, phi=np.pi / 3)
        pq.Q(0) | pq.Displacement(r=0.1, phi=0)
        pq.Q(1) | pq.Displacement(r=0.2, phi=np.pi / 3)
        pq.Q(2) | pq.Displacement(r=0.3, phi=-np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)
        pq.Q(1, 2) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 7)

    simulator = pq.GaussianSimulator(d=3)
    mixed_state = simulator.execute(program).state

    purification = mixed_state.purify()

    assert not mixed_state.is_pure()
    assert purification.is_pure()
    assert purification.d == 2 * mixed_state.d
    assert purification.reduced(modes=(0, 1, 2)) == mixed_state


def test_density_matrix_Thermal_Interferometer(generate_unitary_matrix):
    d = 3

    preparation = pq.Program([pq.Thermal([0.1, 0.2, 0.3])])

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.Interferometer(U)

    simulator = pq.GaussianSimulator(d=d, config=pq.Config(cutoff=2))

    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    fock_space_unitary = block_diag(np.array([1.0]), U)

    assert np.allclose(
        final_state.density_matrix,
        fock_space_unitary @ initial_state.density_matrix @ fock_space_unitary.conj().T,
    )


def test_get_purity_Thermal_Interferometer(generate_unitary_matrix):
    d = 3

    preparation = pq.Program([pq.Thermal([0.1, 0.2, 0.3])])

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.Interferometer(U)

    simulator = pq.GaussianSimulator(d=d, config=pq.Config(cutoff=7))

    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    initial_density_matrix = initial_state.density_matrix
    final_density_matrix = final_state.density_matrix

    actual_initial_purity = np.trace(initial_density_matrix @ initial_density_matrix)
    actual_final_purity = np.trace(final_density_matrix @ final_density_matrix)

    expected_initial_purity = initial_state.get_purity()
    expected_final_purity = final_state.get_purity()

    assert np.isclose(actual_initial_purity, actual_final_purity)
    assert np.isclose(expected_initial_purity, expected_final_purity)
    assert np.isclose(expected_final_purity, actual_final_purity)


def test_get_xp_string_moment_vacuum_state_first_and_second_moments():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    state = pq.GaussianSimulator(d=1).execute(program).state

    assert np.isclose(state.get_xp_string_moment([]), 1.0)

    assert np.isclose(state.get_xp_string_moment([0]), 0.0)
    assert np.isclose(state.get_xp_string_moment([1]), 0.0)

    assert np.isclose(state.get_xp_string_moment([0, 0]), 1.0)
    assert np.isclose(state.get_xp_string_moment([1, 1]), 1.0)
    assert np.isclose(state.get_xp_string_moment([0, 1]), 1.0j)
    assert np.isclose(state.get_xp_string_moment([1, 0]), -1.0j)


def test_get_xp_string_moment_vacuum_energy():
    config = pq.Config()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    state = pq.GaussianSimulator(d=1, config=config).execute(program).state
    energy = (
        state.get_xp_string_moment([0, 0]) + state.get_xp_string_moment([1, 1])
    ) / 2

    assert np.isclose(energy, config.hbar / 2)


def test_get_xp_string_moment_vacuum_state_degree_4_decomposes_into_degree_2():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    state = pq.GaussianSimulator(d=1).execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0, 0, 1, 1]),
        state.get_xp_string_moment([0, 0]) * state.get_xp_string_moment([1, 1])
        + 2 * state.get_xp_string_moment([0, 1]) ** 2,
    )


def test_get_xp_string_moment_vacuum_state_against_explicit_formula():
    r"""
    We know that
    .. math::
        \bra{0} \hat{x}^n \ket{0} = \bra{0} \hat{p}^n \ket{0} = (n-1)!!,

    see, e.g., https://physics.stackexchange.com/a/748168/325715.
    """

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    state = pq.GaussianSimulator(d=1).execute(program).state

    assert np.isclose(state.get_xp_string_moment([0] * 6), factorial2(6 - 1))
    assert np.isclose(state.get_xp_string_moment([0] * 8), factorial2(8 - 1))

    assert np.isclose(state.get_xp_string_moment([1] * 6), factorial2(6 - 1))
    assert np.isclose(state.get_xp_string_moment([1] * 8), factorial2(8 - 1))


def test_get_xp_string_moment_admits_ccr_for_vacuum_state():
    config = pq.Config()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    state = pq.GaussianSimulator(d=1, config=config).execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0, 0, 1, 1])
        - state.get_xp_string_moment([0, 1, 0, 1]),
        1j * config.hbar * state.get_xp_string_moment([0, 1]),
    )


def test_get_xp_string_moment_single_mode_squeezed_state():
    r"""
    Consider :math:`\ket{r} := S(r) \ket{0}`, which is the state vector of a single-mode
    squeezed vacuum.

    Then, we know that
    .. math::
        \bra{r} \hat{X} \ket{r}
            = e^{-(n-m)r} \bra{0} \hat{X} \ket{0},

    where :math:`\hat{X}` is an XP-string consisting of :math:`n` :math:`\hat{x}`
    operators and :math:`m` :math:`\hat{p}` operators.
    """

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    vacuum_state = pq.GaussianSimulator(d=1).execute(program).state

    r = 0.2

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Squeezing(r=r)

    state = pq.GaussianSimulator(d=1).execute(program).state

    xp_string = [0, 1, 0, 0, 1, 1, 0]

    m = np.count_nonzero(xp_string)
    n = len(xp_string) - m

    assert np.isclose(
        state.get_xp_string_moment(xp_string),
        np.exp(-(n - m) * r) * vacuum_state.get_xp_string_moment(xp_string),
    )


def test_get_xp_string_moment_coherent_state_first_and_second_moments():
    r"""
    This test uses the following fact:

    .. math::
        D(r, \phi)^\dagger \hat{x} D(r, \phi) = \hat{x} + \sqrt{2 \hbar} r \cos \phi, \\
        D(r, \phi)^\dagger \hat{p} D(r, \phi) = \hat{x} + \sqrt{2 \hbar} r \sin \phi.
    """

    hbar = 2
    config = pq.Config(hbar=hbar)

    r = 0.2
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=r, phi=phi)

    state = pq.GaussianSimulator(d=1, config=config).execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0]), np.sqrt(2 * hbar) * r * np.cos(phi)
    )
    assert np.isclose(
        state.get_xp_string_moment([1]), np.sqrt(2 * hbar) * r * np.sin(phi)
    )
    assert np.isclose(
        state.get_xp_string_moment([0, 1]),
        hbar * (1j / 2 + 2 * r**2 * np.cos(phi) * np.sin(phi)),
    )

    assert np.isclose(
        state.get_xp_string_moment([0, 0]), hbar * (1 / 2 + 2 * r**2 * np.cos(phi) ** 2)
    )


def test_get_xp_string_moment_coherent_state_higher_moments():
    r"""
    This test uses the following fact:

    .. math::
        D(r, \phi)^\dagger \hat{x} D(r, \phi) &= \hat{x} + \sqrt{2 \hbar} r \cos \phi, \\
        D(r, \phi)^\dagger \hat{p} D(r, \phi) &= \hat{x} + \sqrt{2 \hbar} r \sin \phi.
    """  # noqa: E501

    hbar = 2
    config = pq.Config(hbar=hbar)

    r = 0.2
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=r, phi=phi)

    state = pq.GaussianSimulator(d=1, config=config).execute(program).state

    x_moment = state.get_xp_string_moment([0])
    p_moment = state.get_xp_string_moment([1])

    moment_000 = state.get_xp_string_moment([0, 0, 0])

    assert np.isclose(
        moment_000,
        3 * (state.get_xp_string_moment([0, 0]) - x_moment**2) * x_moment + x_moment**3,
    )
    assert np.isclose(
        moment_000,
        np.sqrt(hbar) ** 3
        * (3 * r * np.cos(phi) / np.sqrt(2) + (np.sqrt(2) * r * np.cos(phi)) ** 3),
    )

    moment_011 = state.get_xp_string_moment([0, 1, 1])

    assert np.isclose(
        moment_011,
        x_moment * (state.get_xp_string_moment([1, 1]) - p_moment**2)
        + 2 * p_moment * (state.get_xp_string_moment([0, 1]) - x_moment * p_moment)
        + x_moment * p_moment**2,
    )

    assert np.isclose(
        moment_011,
        np.sqrt(hbar) ** 3
        * (
            np.sqrt(2**3) * r**3 * np.cos(phi) * np.sin(phi) ** 2
            + r * np.cos(phi) / np.sqrt(2)
            + 1j * np.sqrt(2) * r * np.sin(phi)
        ),
    )


def test_get_xp_string_moment_2_mode_first_and_second_moments():
    r"""
    This test uses the photonic `ControlledZ` gate (:math:`\hbar=1`):

    .. math::
        CZ(s)^\dagger \hat{x}_i CZ(s) &= \hat{x}_i, \\
        CZ(s)^\dagger \hat{p}_1 CZ(s) &= \hat{p}_1 + s \hat{x}_2, \\
        CZ(s)^\dagger \hat{p}_2 CZ(s) &= \hat{p}_2 + s \hat{x}_1,

    From these relations, the moments can be easily calculated.
    """
    config = pq.Config(hbar=1)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0, 1) | pq.ControlledZ(s=1)

    state = pq.GaussianSimulator(d=2, config=config).execute(program).state

    assert np.isclose(state.get_xp_string_moment([]), 1.0)
    assert np.isclose(state.get_xp_string_moment([0]), 0.0)
    assert np.isclose(state.get_xp_string_moment([1]), 0.0)
    assert np.isclose(state.get_xp_string_moment([2]), 0.0)
    assert np.isclose(state.get_xp_string_moment([3]), 0.0)

    second_moment_matrix = np.empty(shape=(4, 4), dtype=config.complex_dtype)

    for i in range(4):
        for j in range(4):
            second_moment_matrix[i, j] = state.get_xp_string_moment([i, j])

    assert np.allclose(
        second_moment_matrix,
        [
            [0.5 + 0.0j, 0.0 + 0.0j, 0.0 + 0.5j, 0.5 + 0.0j],
            [0.0 + 0.0j, 0.5 + 0.0j, 0.5 + 0.0j, 0.0 + 0.5j],
            [0.0 - 0.5j, 0.5 + 0.0j, 1.0 + 0.0j, 0.0 + 0.0j],
            [0.5 + 0.0j, 0.0 - 0.5j, 0.0 + 0.0j, 1.0 + 0.0j],
        ],
    )


def test_get_xp_string_moment_2_mode_higher_moments():
    r"""
    This test uses the photonic `ControlledZ` gate (:math:`\hbar=1`):

    .. math::
        CZ(s)^\dagger \hat{x}_i CZ(s) &= \hat{x}_i, \\
        CZ(s)^\dagger \hat{p}_1 CZ(s) &= \hat{p}_1 + s \hat{x}_2, \\
        CZ(s)^\dagger \hat{p}_2 CZ(s) &= \hat{p}_2 + s \hat{x}_1,

    From these relations, the moments can be easily calculated.
    """
    config = pq.Config(hbar=1)

    with pq.Program() as vacuum_program:
        pq.Q() | pq.Vacuum()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0, 1) | pq.ControlledZ(s=1)

    simulator = pq.GaussianSimulator(d=2, config=config)

    vacuum_state = simulator.execute(vacuum_program).state
    state = simulator.execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0, 2, 0, 2]),
        vacuum_state.get_xp_string_moment([0, 2, 0, 2])
        + vacuum_state.get_xp_string_moment([0, 1, 0, 2])
        + vacuum_state.get_xp_string_moment([0, 2, 0, 1])
        + vacuum_state.get_xp_string_moment([0, 1, 0, 1]),
    )


def test_get_xp_string_moment_admits_ccr_1():
    config = pq.Config()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Displacement(r=0.4)
        pq.Q(0) | pq.Squeezing(r=0.2, phi=0.3)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5)

    state = pq.GaussianSimulator(d=2, config=config).execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0, 0, 2, 2])
        - state.get_xp_string_moment([0, 2, 0, 2]),
        1j * config.hbar * state.get_xp_string_moment([0, 2]),
    )


def test_get_xp_string_moment_admits_ccr_2():
    config = pq.Config()

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Displacement(r=0.4)
        pq.Q(0) | pq.Squeezing(r=0.2, phi=0.3)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5)

    state = pq.GaussianSimulator(d=2, config=config).execute(program).state

    assert np.isclose(
        state.get_xp_string_moment([0, 1, 2, 3]),
        state.get_xp_string_moment([0, 2, 1, 3]),
    )


def test_get_xp_string_moment_hbar_dependence():
    hbar_1 = 1.0
    hbar_2 = 3.0

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Displacement(r=0.4)
        pq.Q(0) | pq.Squeezing(r=0.2, phi=0.3)
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5)

    state1 = (
        pq.GaussianSimulator(d=2, config=pq.Config(hbar=hbar_1)).execute(program).state
    )
    state2 = (
        pq.GaussianSimulator(d=2, config=pq.Config(hbar=hbar_2)).execute(program).state
    )

    indices = [0, 2, 1, 3, 0, 2]

    value1 = state1.get_xp_string_moment(indices)
    value2 = state2.get_xp_string_moment(indices)

    assert np.isclose(value1 * np.sqrt(hbar_2 / hbar_1) ** len(indices), value2)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from piquasso.fermionic._utils import binary_to_fock_indices

for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@pytest.mark.monkey
@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_equivalence_1_particle(
    connector,
    generate_unitary_matrix,
):
    d = 3

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 1])

        pq.Q() | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix
    density_matrix_fock = fock_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(
        density_matrix_fock[1 : (d + 1), 1 : (d + 1)],
        U @ np.diag([0, 0, 1]) @ U.conj().T,
    )

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@pytest.mark.monkey
@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_equivalence_2_particles(
    connector,
    generate_unitary_matrix,
):
    d = 3

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 0, 1])

        pq.Q() | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_fock = fock_simulator.execute(program).state.density_matrix
    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@pytest.mark.monkey
@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_equivalence_3_particles(
    connector,
    generate_unitary_matrix,
):
    d = 4

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 0, 1, 1])

        pq.Q() | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_fock = fock_simulator.execute(program).state.density_matrix
    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@pytest.mark.monkey
@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_equivalence_n_particles_random(
    connector,
    generate_unitary_matrix,
):
    d = np.random.randint(1, 6)

    occupation_numbers = np.random.randint(0, 2, d)

    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers)

        pq.Q() | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_fock = fock_simulator.execute(program).state.density_matrix
    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@pytest.mark.monkey
@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_equivalence_subsystem_random(
    connector,
    generate_unitary_matrix,
):
    d = 3

    occupation_numbers = [1, 1, 0]

    modes = (0, 1)

    U = generate_unitary_matrix(2)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers)

        pq.Q(*modes) | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_fock = fock_simulator.execute(program).state.density_matrix
    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_squeezing2_equivalence(connector):
    d = 2

    r = 0.2
    phi = np.pi / 5

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1])

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix
    density_matrix_fock = fock_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)


@for_all_connectors
def test_PureFockSimulator_GaussianSimulator_IsingXX_equivalence(connector):
    d = 2

    phi = np.pi / 5

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1])

        pq.Q(0, 1) | pq.fermionic.IsingXX(phi=phi)

    fock_simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )
    gaussian_simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    density_matrix_gaussian = gaussian_simulator.execute(program).state.density_matrix
    density_matrix_fock = fock_simulator.execute(program).state.density_matrix

    indices = binary_to_fock_indices(d)

    density_matrix_gaussian_reordered = density_matrix_gaussian[
        np.ix_(indices, indices)
    ]

    assert np.allclose(density_matrix_fock, density_matrix_gaussian_reordered)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some tools for simulating CVQNN (Continuous-Variable Quantum Neural Networks) that are
available in Piquasso.
"""

from typing import List, Optional

import numpy as np

from dataclasses import dataclass

from piquasso.api.program import Program
from piquasso.api.instruction import Instruction
from piquasso.api.exceptions import CVQNNException
from piquasso.instructions.preparations import Vacuum
from piquasso.instructions.gates import (
    Phaseshifter,
    Beamsplitter,
    Squeezing,
    Displacement,
    Kerr,
)


@dataclass
class _InterferometerParameters:
    thetas: np.ndarray
    phis: np.ndarray
    final_phis: np.ndarray


@dataclass
class _LayerParameters:
    first_interferometer: _InterferometerParameters
    squeezing_parameters: np.ndarray
    second_interferometer: _InterferometerParameters
    displacement_amplitudes: np.ndarray
    displacement_angles: np.ndarray
    kerr_parameters: np.ndarray


def get_cvqnn_weight_indices(d):
    M = _get_number_of_interferometer_parameters(d)

    K = _get_number_of_single_layer_parameters(d)

    index_range = np.arange(K)

    return [
        index_range[:M],
        index_range[M : M + d],
        index_range[M + d : 2 * M + d],
        index_range[2 * M + d : 2 * M + 2 * d],
        index_range[2 * M + 2 * d : 2 * M + 3 * d],
        index_range[2 * M + 3 * d :],
    ]


def generate_random_cvqnn_weights(
    layer_count: int,
    d: int,
    active_var: float = 0.01,
    passive_var: float = 0.1,
    rng: Optional[np.random.Generator] = None,
) -> np.ndarray:
    """Generates random CVQNN weights for :func:`create_program`.

    Args:
        layer_count (int): The number of layers.
        d (int): Number of modes.
        active_var (float, optional): Active gate variance. Defaults to 0.01.
        passive_var (float, optional): Passive gate variance. Defaults to 0.1.

    Returns:
        np.ndarray: An array of weights.
    """
    if rng is None:
        rng = np.random.default_rng()

    M = _get_number_of_interferometer_parameters(d)

    first_interferometer = rng.normal(size=[layer_count, M], scale=passive_var)
    squeezing_parameters = rng.normal(size=[layer_count, d], scale=active_var)
    second_interferometer = rng.normal(size=[layer_count, M], scale=passive_var)
    displacement_amplitudes = rng.normal(size=[layer_count, d], scale=active_var)
    displacement_angles = rng.normal(size=[layer_count, d], scale=passive_var)
    kerr_parameters = rng.normal(size=[layer_count, d], scale=active_var)

    return np.concatenate(
        [
            first_interferometer,
            squeezing_parameters,
            second_interferometer,
            displacement_amplitudes,
            displacement_angles,
            kerr_parameters,
        ],
        axis=1,
    )


def create_layers(weights: np.ndarray) -> Program:
    """Creates a subprogram from the specified weights.

    Example usage::

        import piquasso as pq

        weights = pq.cvqnn.generate_random_cvqnn_weights(layer_count=10, d=5)

        cvnn_layers = pq.cvqnn.create_program(weights)

        with pq.Program() as program:
            for i in range(5):
                pq.Q(i) | pq.Displacement(r=0.1)

            pq.Q() | cvnn_layers

    Args:
        weights (np.ndarray): The CVQNN circuit weights.

    Returns:
        Program: The program of the CVQNN circuit, ready to be executed.
    """

    d = get_number_of_modes(weights.shape[1])
    layer_count = weights.shape[0]

    instructions = []

    for k in range(layer_count):
        layer_parameters = _parse_layer(weights[k], d)
        instructions.extend(_instructions_from_layer(layer_parameters, d))

    return Program(instructions)


def create_program(weights: np.ndarray) -> Program:
    """Creates a `Program` from the specified weights, with vacuum initial state.

    Args:
        weights (np.ndarray): The CVQNN circuit weights.

    Returns:
        Program: The program of the CVQNN circuit, ready to be executed.
    """

    cvqnn_layers = create_layers(weights)

    program = Program(instructions=[Vacuum()] + cvqnn_layers.instructions)

    return program


def _get_number_of_single_layer_parameters(d):
    return 2 * _get_number_of_interferometer_parameters(d) + 4 * d


def get_number_of_modes(number_of_parameters: int) -> int:
    """Returns the number of modes given the number of parameters in a CVQNN layer.

    Args:
        number_of_parameters (int): The number of CVQNN parameters per layer.

    Raises:
        CVQNNException:
            If the number of parameters does not correspond to any number of modes

    Returns:
        int: The number of modes.
    """
    if number_of_parameters == 6:
        return 1

    d = 2

    while _get_number_of_single_layer_parameters(d) < number_of_parameters:
        d += 1

    if not _get_number_of_single_layer_parameters(d) == number_of_parameters:
        raise CVQNNException(
            f"Invalid number of parameters specified: '{number_of_parameters}'."
        )

    return d


def _get_number_of_interferometer_parameters(d):
    return int(d * (d - 1)) + max(1, d - 1)


def _parse_interferometer(weights: np.ndarray, d: int) -> _InterferometerParameters:
    return _InterferometerParameters(
        thetas=weights[: d * (d - 1) // 2],
        phis=weights[d * (d - 1) // 2 : d * (d - 1)],
        final_phis=weights[-d + 1 :],
    )


def _parse_layer(weights: np.ndarray, d: int) -> _LayerParameters:
    M = _get_number_of_interferometer_parameters(d)

    return _LayerParameters(
        _parse_interferometer(weights[:M], d),
        weights[M : M + d],
        _parse_interferometer(weights[M + d : 2 * M + d], d),
        weights[2 * M + d : 2 * M + 2 * d],
        weights[2 * M + 2 * d : 2 * M + 3 * d],
        weights[2 * M + 3 * d : 2 * M + 4 * d],
    )


def _instructions_from_interferometer(
    interferometer_parameters: _InterferometerParameters, d: int
) -> List[Instruction]:
    instructions = []

    thetas = interferometer_parameters.thetas
    phis = interferometer_parameters.phis
    final_phis = interferometer_parameters.final_phis

    if d == 1:
        return [Phaseshifter(final_phis[0]).on_modes(0)]

    n = 0
    indices = [list(range(0, d - 1, 2)), list(range(1, d - 1, 2))]

    for j in range(d):
        for k in indices[j % 2]:
            instructions.append(
                Beamsplitter(theta=thetas[n], phi=phis[n]).on_modes(k, k + 1)
            )
            n += 1

    instructions.extend(
        [Phaseshifter(final_phis[i]).on_modes(i) for i in range(max(1, d - 1))]
    )

    return instructions


def _instructions_from_layer(
    layer_parameters: _LayerParameters, d: int
) -> List[Instruction]:
    instructions = []

    instructions.extend(
        _instructions_from_interferometer(layer_parameters.first_interferometer, d)
    )

    squeezing_parameters = layer_parameters.squeezing_parameters
    instructions.extend(
        [Squeezing(squeezing_parameters[i]).on_modes(i) for i in range(d)]
    )

    instructions.extend(
        _instructions_from_interferometer(layer_parameters.second_interferometer, d)
    )

    displacement_amplitudes = layer_parameters.displacement_amplitudes
    displacement_angles = layer_parameters.displacement_angles
    instructions.extend(
        [
            Displacement(displacement_amplitudes[i], displacement_angles[i]).on_modes(i)
            for i in range(d)
        ]
    )

    kerr_parameters = layer_parameters.kerr_parameters
    instructions.extend([Kerr(kerr_parameters[i]).on_modes(i) for i in range(d)])

    return instructions



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.instructions import gates, preparations
from .state import PureFockState

from .calculations import (
    passive_linear,
    state_vector,
    squeezing2,
    controlled_phase,
    ising_XX,
)

from piquasso._simulators.simulator import BuiltinSimulator
from piquasso._simulators.connectors import NumpyConnector, JaxConnector
from piquasso.fermionic.instructions import ControlledPhase, IsingXX


class PureFockSimulator(BuiltinSimulator):
    """
    Fermionic pure Fock state simulator.

    Example usage::

        U = scipy.stats.unitary_group.rvs(4)

        with pq.Program() as program:
            pq.Q() | pq.StateVector([0, 0, 1, 1]) / np.sqrt(2)
            pq.Q() | pq.StateVector([1, 1, 0, 0]) / np.sqrt(2)
            pq.Q() | pq.Interferometer(U)

        simulator = pq.fermionic.PureFockSimulator(
            d=4, config=pq.Config(cutoff=4 + 1)
        )

        state = simulator.execute(program).state

    Supported preparations:
        :class:`~piquasso.instructions.preparations.Vacuum`,
        :class:`~piquasso.instructions.preparations.StateVector`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`.
    """

    _state_class = PureFockState

    _instruction_map = {
        preparations.StateVector: state_vector,
        gates.Interferometer: passive_linear,
        gates.Phaseshifter: passive_linear,
        gates.Beamsplitter: passive_linear,
        gates.Beamsplitter5050: passive_linear,
        gates.MachZehnder: passive_linear,
        gates.Fourier: passive_linear,
        gates.Squeezing2: squeezing2,
        ControlledPhase: controlled_phase,
        IsingXX: ising_XX,
    }

    _default_connector_class = NumpyConnector

    _extra_builtin_connectors = [JaxConnector]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

from piquasso.api.result import Result


def test_Result_repr(FakeState):
    result = Result(
        state=FakeState(d=3, connector=pq.NumpyConnector()), samples=[1, 2, 3]
    )

    assert (
        repr(result)
        == "Result(samples=[1, 2, 3], state=FakeState(d=3, config=Config(), connector=NumpyConnector()))"  # noqa: E501
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt

import piquasso as pq


def main():
    d = 5
    shots = 10000
    measurement_cutoff = 6

    # NOTE: In SF the measurement cutoff is 5, couldn't be changed, and in PQ it
    # corresponds to `measurement_cutoff=6`.
    simulator = pq.GaussianSimulator(
        d=d, config=pq.Config(measurement_cutoff=measurement_cutoff)
    )

    with pq.Program() as program:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

        pq.Q(all) | pq.ThresholdMeasurement()

    result = simulator.execute(program, shots=shots)

    samples = result.samples
    state = result.state

    samples_as_list = [tuple(x) for x in samples]

    sample_set = set(samples_as_list)

    f_obs = []
    f_exp = []

    for sample in sample_set:
        count = samples_as_list.count(sample)
        if count < 10:
            continue

        f_obs.append(count / shots)
        f_exp.append(state.get_threshold_detection_probability(sample).real)

    f_obs = np.array(f_obs) / np.sum(f_obs)
    f_exp = np.array(f_exp) / np.sum(f_exp)

    plt.bar(np.arange(len(f_obs)), f_exp, fc=(0.5, 0, 0.5, 0.5), label="Expected")
    plt.bar(np.arange(len(f_obs)), f_obs, fc=(0, 1, 0, 0.5), label="Observed")

    plt.xlim(-1, len(f_obs))

    plt.legend()

    plt.get_current_fig_manager().full_screen_toggle()

    plt.show()


if __name__ == "__main__":
    main()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .state import PureFockState

from .calculations import (
    state_vector_instruction,
    passive_linear,
    beamsplitter5050,
    kerr,
    cross_kerr,
    cubic_phase,
    squeezing,
    displacement,
    linear,
    particle_number_measurement,
    vacuum,
    create,
    annihilate,
    batch_prepare,
    batch_apply,
    post_select_photons,
    imperfect_post_select_photons,
    homodyne_measurement,
)

from ..calculations import attenuator

from ...simulator import BuiltinSimulator
from piquasso.instructions import (
    preparations,
    gates,
    measurements,
    channels,
    batch,
)

from piquasso._simulators.connectors import (
    NumpyConnector,
    TensorflowConnector,
    JaxConnector,
)


class PureFockSimulator(BuiltinSimulator):
    """Performs photonic simulations using Fock representation with pure states.

    The simulation (when executed) results in an instance of
    :class:`~piquasso._simulators.fock.pure.state.PureFockState`.

    Example usage::

        import numpy as np
        import piquasso as pq


        with pq.Program() as program:
            pq.Q(all) | pq.Vacuum()

            pq.Q(0)   | pq.Squeezing(r=0.1)
            pq.Q(1)   | pq.Squeezing(r=0.2)

            pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3)

            pq.Q(0) | pq.Kerr(xi=0.05)

        simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=7))
        result = simulator.execute(program)

    Supported preparations:
        :class:`~piquasso.instructions.preparations.Vacuum`,
        :class:`~piquasso.instructions.preparations.Create`,
        :class:`~piquasso.instructions.preparations.Annihilate`,
        :class:`~piquasso.instructions.preparations.StateVector`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`,
        :class:`~piquasso.instructions.gates.Beamsplitter`,
        :class:`~piquasso.instructions.gates.Phaseshifter`,
        :class:`~piquasso.instructions.gates.MachZehnder`,
        :class:`~piquasso.instructions.gates.Fourier`,
        :class:`~piquasso.instructions.gates.Kerr`,
        :class:`~piquasso.instructions.gates.CrossKerr`,
        :class:`~piquasso.instructions.gates.CubicPhase`,
        :class:`~piquasso.instructions.gates.GaussianTransform`,
        :class:`~piquasso.instructions.gates.Squeezing`,
        :class:`~piquasso.instructions.gates.QuadraticPhase`,
        :class:`~piquasso.instructions.gates.Squeezing2`,
        :class:`~piquasso.instructions.gates.ControlledX`,
        :class:`~piquasso.instructions.gates.ControlledZ`,
        :class:`~piquasso.instructions.gates.Displacement`,
        :class:`~piquasso.instructions.gates.PositionDisplacement`,
        :class:`~piquasso.instructions.gates.MomentumDisplacement`.

    Supported measurements:
        :class:`~piquasso.instructions.measurements.ParticleNumberMeasurement`.

    Supported channels:
        :class:`~piquasso.instructions.channels.Attenuator`.
    """

    _state_class = PureFockState

    _instruction_map = {
        preparations.Vacuum: vacuum,
        preparations.Create: create,
        preparations.Annihilate: annihilate,
        preparations.StateVector: state_vector_instruction,
        gates.Interferometer: passive_linear,
        gates.Beamsplitter: passive_linear,
        gates.Beamsplitter5050: beamsplitter5050,
        gates.Phaseshifter: passive_linear,
        gates.MachZehnder: passive_linear,
        gates.Fourier: passive_linear,
        gates.Kerr: kerr,
        gates.CrossKerr: cross_kerr,
        gates.CubicPhase: cubic_phase,
        gates.Squeezing: squeezing,
        gates.QuadraticPhase: linear,
        gates.Displacement: displacement,
        gates.PositionDisplacement: displacement,
        gates.MomentumDisplacement: displacement,
        gates.Squeezing2: linear,
        gates.GaussianTransform: linear,
        measurements.ParticleNumberMeasurement: particle_number_measurement,
        measurements.PostSelectPhotons: post_select_photons,
        measurements.ImperfectPostSelectPhotons: imperfect_post_select_photons,
        measurements.HomodyneMeasurement: homodyne_measurement,
        channels.Attenuator: attenuator,
        batch.BatchPrepare: batch_prepare,
        batch.BatchApply: batch_apply,
    }

    _default_connector_class = NumpyConnector

    _extra_builtin_connectors = [TensorflowConnector, JaxConnector]
